#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The repository API of the nomad@FAIRDI APIs. Currently allows to resolve repository
meta-data.
'''

from typing import List, Dict, Any
from flask_restplus import Resource, abort, fields
from flask import request, g
from elasticsearch_dsl import Q
from elasticsearch.exceptions import NotFoundError
import elasticsearch.helpers
from datetime import datetime

from nomad import search, utils, datamodel, processing as proc, infrastructure, files, metainfo
from nomad.datamodel import Dataset, User, EditableUserMetadata

from .. import common
from ..common import RFC3339DateTime, DotKeyNested
from .api import api
from .auth import authenticate
from .common import search_model, calc_route, add_pagination_parameters,\
    add_scroll_parameters, add_search_parameters, apply_search_parameters,\
    query_api_python, query_api_curl, query_api_clientlib, query_api_repo_url, \
    _search_quantities

ns = api.namespace('repo', description='Access repository metadata.')


@calc_route(ns)
class RepoCalcResource(Resource):
    @api.response(404, 'The upload or calculation does not exist')
    @api.response(401, 'Not authorized to access the calculation')
    @api.response(200, 'Metadata send', fields.Raw)
    @api.doc('get_repo_calc')
    @authenticate()
    def get(self, upload_id, calc_id):
        '''
        Get calculation metadata in repository form.

        Repository metadata only entails the quantities shown in the repository.
        Calcs are references via *upload_id*, *calc_id* pairs.
        '''
        try:
            calc = search.entry_document.get(calc_id)
        except NotFoundError:
            abort(404, message='There is no calculation %s/%s' % (upload_id, calc_id))

        if calc.with_embargo or not calc.published:
            if g.user is None:
                abort(401, message='Not logged in to access %s/%s.' % (upload_id, calc_id))

            if not (any(g.user.user_id == user.user_id for user in calc.owners) or g.user.is_admin):
                abort(401, message='Not authorized to access %s/%s.' % (upload_id, calc_id))

        result = search._es_to_entry_dict(calc, required=None)
        result['code'] = {
            'python': query_api_python(dict(upload_id=upload_id, calc_id=calc_id)),
            'curl': query_api_curl(dict(upload_id=upload_id, calc_id=calc_id)),
            'clientlib': query_api_clientlib(upload_id=[upload_id], calc_id=[calc_id])
        }

        return result, 200


_search_request_parser = api.parser()
add_pagination_parameters(_search_request_parser)
add_scroll_parameters(_search_request_parser)
add_search_parameters(_search_request_parser)
_search_request_parser.add_argument(
    'date_histogram', type=bool, help='Add an additional aggregation over the upload time')
_search_request_parser.add_argument(
    'interval', type=str, help='Interval to use for upload time aggregation.')
_search_request_parser.add_argument(
    'metrics', type=str, action='append', help=(
        'Metrics to aggregate over all quantities and their values as comma separated list. '
        'Possible values are %s.' % ', '.join(search.metrics.keys())))
_search_request_parser.add_argument(
    'statistics', type=str, action='append', help=(
        'Quantities for which to aggregate values and their metrics.'))
_search_request_parser.add_argument(
    'exclude', type=str, action='split', help='Excludes the given keys in the returned data.')
for group_name in search.groups:
    _search_request_parser.add_argument(
        group_name, type=bool, help=('Return %s group data.' % group_name))
    _search_request_parser.add_argument(
        '%s_after' % group_name, type=str,
        help='The last %s id of the last scroll window for the %s group' % (group_name, group_name))

_repo_calcs_model_fields = {
    'statistics': fields.Raw(description=(
        'A dict with all statistics. Each statistic is dictionary with a metrics dict as '
        'value and quantity value as key. The possible metrics are code runs(calcs), %s. '
        'There is a pseudo quantity "total" with a single value "all" that contains the '
        ' metrics over all results. ' % ', '.join(search.metrics.keys())))}

for group_name in search.groups:
    _repo_calcs_model_fields[group_name] = (DotKeyNested if '.' in group_name else fields.Nested)(api.model('RepoGroup', {
        'after': fields.String(description='The after value that can be used to retrieve the next %s.' % group_name),
        'values': fields.Raw(description='A dict with %s as key. The values are dicts with "total" and "examples" keys.' % group_name)
    }), skip_none=True)

for qualified_name, quantity in search.search_quantities.items():
    _repo_calcs_model_fields[qualified_name] = fields.Raw(
        description=quantity.description, allow_null=True, skip_none=True)

_repo_calcs_model_fields.update(**{
    'date_histogram': fields.Boolean(default=False, description='Add an additional aggregation over the upload time', allow_null=True, skip_none=True),
    'interval': fields.String(description='Interval to use for upload time aggregation.', allow_null=True, skip_none=True),
    'metrics': fields.List(fields.String, description=(
        'Metrics to aggregate over all quantities and their values as comma separated list. '
        'Possible values are %s.' % ', '.join(search.metrics.keys())), allow_null=True, skip_none=True),
    'statistics_required': fields.List(fields.String, description='Quantities for which to aggregate values and their metrics.', allow_null=True, skip_none=True),
    'exclude': fields.List(fields.String, description='Excludes the given keys in the returned data.', allow_null=True, skip_none=True)
})

_repo_calcs_model = api.inherit('RepoCalculations', search_model, _repo_calcs_model_fields)


@ns.route('/')
class RepoCalcsResource(Resource):
    @api.doc('search')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.expect(_search_request_parser, validate=True)
    @api.marshal_with(_repo_calcs_model, skip_none=True, code=200, description='Search results send')
    @authenticate()
    def get(self):
        '''
        Search for calculations in the repository form, paginated.

        The ``owner`` parameter determines the overall entries to search through.
        Possible values are: ``all`` (show all entries visible to the current user), ``public``
        (show all publically visible entries), ``user`` (show all user entries, requires login),
        ``staging`` (show all user entries in staging area, requires login).

        You can use the various quantities to search/filter for. For some of the
        indexed quantities this endpoint returns aggregation information. This means
        you will be given a list of all possible values and the number of entries
        that have the certain value. You can also use these aggregations on an empty
        search to determine the possible values.

        The pagination parameters allows determine which page to return via the
        ``page`` and ``per_page`` parameters. Pagination however, is limited to the first
        100k (depending on ES configuration) hits.

        An alternative to pagination is to use ``scroll`` and ``scroll_id``. With ``scroll``
        you will get a ``scroll_id`` on the first request. Each call with ``scroll`` and
        the respective ``scroll_id`` will return the next ``per_page`` (here the default is 1000)
        results. Scroll however, ignores ordering and does not return aggregations.
        The scroll view used in the background will stay alive for 1 minute between requests.
        If the given ``scroll_id`` is not available anymore, a HTTP 400 is raised.

        The search will return aggregations on a predefined set of quantities. Aggregations
        will tell you what quantity values exist and how many entries match those values.

        Ordering is determined by ``order_by`` and ``order`` parameters. Default is
        ``upload_time`` in decending order.
        '''

        try:
            parsed_args = _search_request_parser.parse_args()
            args = {
                key: value for key, value in parsed_args.items()
                if value is not None}

            scroll = args.get('scroll', False)
            scroll_id = args.get('scroll_id', None)
            page = args.get('page', 1)
            per_page = args.get('per_page', 10 if not scroll else 1000)
            order = args.get('order', -1)
            order_by = args.get('order_by', 'upload_time')

            date_histogram = args.get('date_histogram', False)
            interval = args.get('interval', '1M')
            metrics: List[str] = request.args.getlist('metrics')
            statistics = args.get('statistics', [])
        except Exception as e:
            abort(400, message='bad parameters: %s' % str(e))

        for metric in metrics:
            if metric not in search.metrics:
                abort(400, message='there is no metric %s' % metric)

        search_request = search.SearchRequest()
        apply_search_parameters(search_request, args)
        if date_histogram:
            search_request.date_histogram(interval=interval, metrics_to_use=metrics)

        try:
            assert page >= 1
            assert per_page >= 0
        except AssertionError:
            abort(400, message='invalid pagination')

        if order not in [-1, 1]:
            abort(400, message='invalid pagination')

        if len(statistics) > 0:
            search_request.statistics(statistics, metrics_to_use=metrics)

        group_metrics = [
            group_quantity.metric_name
            for group_name, group_quantity in search.groups.items()
            if args.get(group_name, False)]
        total_metrics = metrics + group_metrics
        if len(total_metrics) > 0:
            search_request.totals(metrics_to_use=total_metrics)

        if 'exclude' in parsed_args:
            excludes = parsed_args['exclude']
            if excludes is not None:
                search_request.exclude(*excludes)

        try:
            if scroll:
                results = search_request.execute_scrolled(scroll_id=scroll_id, size=per_page)

            else:
                for group_name, group_quantity in search.groups.items():
                    if args.get(group_name, False):
                        kwargs: Dict[str, Any] = {}
                        if group_name == 'uploads_grouped':
                            kwargs.update(order_by='upload_time', order='desc')
                        search_request.quantity(
                            group_quantity.qualified_name, size=per_page, examples=1,
                            after=request.args.get('%s_after' % group_name, None),
                            **kwargs)

                results = search_request.execute_paginated(
                    per_page=per_page, page=page, order=order, order_by=order_by)

                # TODO just a work around to make things prettier
                if 'statistics' in results:
                    statistics = results['statistics']
                    if 'code_name' in statistics and 'currupted mainfile' in statistics['code_name']:
                        del(statistics['code_name']['currupted mainfile'])

                if 'quantities' in results:
                    quantities = results.pop('quantities')

                for group_name, group_quantity in search.groups.items():
                    if args.get(group_name, False):
                        results[group_name] = quantities[group_quantity.qualified_name]

            # build python code/curl snippet
            code_args = request.args.to_dict(flat=False)
            if 'statistics' in code_args:
                del(code_args['statistics'])
            results['code'] = {
                'repo_url': query_api_repo_url(code_args),
                'curl': query_api_curl(code_args),
                'python': query_api_python(code_args),
                'clientlib': query_api_clientlib(**code_args)
            }

            return results, 200
        except search.ScrollIdNotFound:
            abort(400, 'The given scroll_id does not exist.')
        except KeyError as e:
            import traceback
            traceback.print_exc()
            abort(400, str(e))

    @api.doc('post_search')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.expect(_repo_calcs_model)
    @api.marshal_with(_repo_calcs_model, skip_none=True, code=200, description='Search results send')
    @authenticate()
    def post(self):
        '''
        Search for calculations in the repository form, paginated.

        The ``owner`` parameter determines the overall entries to search through.
        Possible values are: ``all`` (show all entries visible to the current user), ``public``
        (show all publically visible entries), ``user`` (show all user entries, requires login),
        ``staging`` (show all user entries in staging area, requires login).

        You can use the various quantities to search/filter for. For some of the
        indexed quantities this endpoint returns aggregation information. This means
        you will be given a list of all possible values and the number of entries
        that have the certain value. You can also use these aggregations on an empty
        search to determine the possible values.

        The pagination parameters allows determine which page to return via the
        ``page`` and ``per_page`` parameters. Pagination however, is limited to the first
        100k (depending on ES configuration) hits.

        An alternative to pagination is to use ``scroll`` and ``scroll_id``. With ``scroll``
        you will get a ``scroll_id`` on the first request. Each call with ``scroll`` and
        the respective ``scroll_id`` will return the next ``per_page`` (here the default is 1000)
        results. Scroll however, ignores ordering and does not return aggregations.
        The scroll view used in the background will stay alive for 1 minute between requests.
        If the given ``scroll_id`` is not available anymore, a HTTP 400 is raised.

        The search will return aggregations on a predefined set of quantities. Aggregations
        will tell you what quantity values exist and how many entries match those values.

        Ordering is determined by ``order_by`` and ``order`` parameters. Default is
        ``upload_time`` in decending order.
        '''
        try:
            data_in = request.get_json()

            Scroll = data_in.get('scroll', {})
            scroll = Scroll.get('scroll', False)
            scroll_id = Scroll.get('scroll_id', None)
            pagination = data_in.get('pagination', {})
            page = pagination.get('page', 1)
            per_page = pagination.get('per_page', 10 if not scroll else 1000)
            order = pagination.get('order', -1)
            order_by = pagination.get('order_by', 'upload_time')

            date_histogram = data_in.get('date_histogram', False)
            interval = data_in.get('interval', '1M')
            metrics: List[str] = data_in.get('metrics', [])
            statistics = data_in.get('statistics_required', [])

            query = data_in.get('query', {})
            query_expression = {key: val for key, val in query.items() if '$' in key}
        except Exception as e:
            abort(400, message='bad parameters: %s' % str(e))

        for metric in metrics:
            if metric not in search.metrics:
                abort(400, message='there is no metric %s' % metric)

        search_request = search.SearchRequest()
        apply_search_parameters(search_request, query)
        if date_histogram:
            search_request.date_histogram(interval=interval, metrics_to_use=metrics)

        if query_expression:
            try:
                search_request.query_expression(query_expression)
            except AssertionError as e:
                abort(400, str(e))

        try:
            assert page >= 1
            assert per_page >= 0
        except AssertionError:
            abort(400, message='invalid pagination')

        if order not in [-1, 1]:
            abort(400, message='invalid pagination')

        if len(statistics) > 0:
            search_request.statistics(statistics, metrics_to_use=metrics)

        group_metrics = [
            group_quantity.metric_name
            for group_name, group_quantity in search.groups.items()
            if group_name in data_in]
        total_metrics = metrics + group_metrics
        if len(total_metrics) > 0:
            search_request.totals(metrics_to_use=total_metrics)

        if 'exclude' in data_in:
            excludes = data_in['exclude']
            if excludes is not None:
                search_request.exclude(*excludes)

        try:
            if scroll:
                results = search_request.execute_scrolled(scroll_id=scroll_id, size=per_page)

            else:
                for group_name, group_quantity in search.groups.items():
                    if group_name in data_in:
                        kwargs: Dict[str, Any] = {}
                        if group_name == 'uploads_grouped':
                            kwargs.update(order_by='upload_time', order='desc')
                        search_request.quantity(
                            group_quantity.qualified_name, size=per_page, examples=1,
                            after=data_in[group_name].get('after', None),
                            **kwargs)

                results = search_request.execute_paginated(
                    per_page=per_page, page=page, order=order, order_by=order_by)

                # TODO just a work around to make things prettier
                if 'statistics' in results:
                    statistics = results['statistics']
                    if 'code_name' in statistics and 'currupted mainfile' in statistics['code_name']:
                        del(statistics['code_name']['currupted mainfile'])

                if 'quantities' in results:
                    quantities = results.pop('quantities')

                for group_name, group_quantity in search.groups.items():
                    if group_name in data_in:
                        results[group_name] = quantities[group_quantity.qualified_name]

            # build python code/curl snippet
            code_args = dict(data_in)
            if 'statistics' in code_args:
                del(code_args['statistics'])
            results['code'] = {
                'curl': query_api_curl(code_args),
                'python': query_api_python(code_args),
                'clientlib': query_api_clientlib(**code_args)
            }

            return results, 200
        except search.ScrollIdNotFound:
            abort(400, 'The given scroll_id does not exist.')
        except KeyError as e:
            import traceback
            traceback.print_exc()
            abort(400, str(e))


_query_model_parameters = {
    'owner': fields.String(description='Specify which calcs to return: ``all``, ``public``, ``user``, ``staging``, default is ``all``'),
    'from_time': RFC3339DateTime(description='A yyyy-MM-ddTHH:mm:ss (RFC3339) minimum entry time (e.g. upload time)'),
    'until_time': RFC3339DateTime(description='A yyyy-MM-ddTHH:mm:ss (RFC3339) maximum entry time (e.g. upload time)')
}

for qualified_name, quantity in search.search_quantities.items():
    if quantity.many_and == 'append' or quantity.many_or == 'append':
        def field(**kwargs):
            return fields.List(fields.String(**kwargs))
    else:
        field = fields.String
    _query_model_parameters[qualified_name] = field(description=quantity.description)

_repo_query_model = api.model('RepoQuery', _query_model_parameters, skip_none=True)


def repo_edit_action_field(quantity):
    if quantity.is_scalar:
        return fields.Nested(_repo_edit_action_model, description=quantity.description, skip_none=True)
    else:
        return fields.List(
            fields.Nested(_repo_edit_action_model, skip_none=True), description=quantity.description)


_repo_edit_action_model = api.model('RepoEditAction', {
    'value': fields.String(description='The value/values that is set as a string.'),
    'success': fields.Boolean(description='If this can/could be done. Only in API response.'),
    'message': fields.String(descriptin='A message that details the action result. Only in API response.')
})

_repo_edit_model = api.model('RepoEdit', {
    'verify': fields.Boolean(description='If true, no action is performed.'),
    'query': fields.Nested(_repo_query_model, skip_none=True, description='New metadata will be applied to query results.'),
    'actions': fields.Nested(
        api.model('RepoEditActions', {
            quantity.name: repo_edit_action_field(quantity)
            for quantity in EditableUserMetadata.m_def.definitions
            if isinstance(quantity, metainfo.Quantity)
        }), skip_none=True,
        description='Each action specifies a single value (even for multi valued quantities).'),
    'success': fields.Boolean(description='If the overall edit can/could be done. Only in API response.'),
    'message': fields.String(description='A message that details the overall edit result. Only in API response.')
})

_editable_quantities = {
    quantity.name: quantity for quantity in EditableUserMetadata.m_def.definitions}


def edit(parsed_query: Dict[str, Any], mongo_update: Dict[str, Any] = None, re_index=True) -> List[str]:
    # get all calculations that have to change
    with utils.timer(common.logger, 'edit query executed'):
        search_request = search.SearchRequest().include('calc_id', 'upload_id')
        apply_search_parameters(search_request, parsed_query)
        upload_ids = set()
        calc_ids = []

        for hit in search_request.execute_scan():
            calc_ids.append(hit['calc_id'])
            upload_ids.add(hit['upload_id'])

    # perform the update on the mongo db
    with utils.timer(common.logger, 'edit mongo update executed', size=len(calc_ids)):
        if mongo_update is not None:
            n_updated = proc.Calc.objects(calc_id__in=calc_ids).update(multi=True, **mongo_update)
            if n_updated != len(calc_ids):
                common.logger.error('edit repo did not update all entries', payload=mongo_update)

    # re-index the affected entries in elastic search
    with utils.timer(common.logger, 'edit elastic update executed', size=len(calc_ids)):
        if re_index:
            def elastic_updates():
                upload_files_cache: Dict[str, files.UploadFiles] = dict()

                for calc in proc.Calc.objects(calc_id__in=calc_ids):
                    upload_id = calc.upload_id
                    upload_files = upload_files_cache.get(upload_id)
                    if upload_files is None:
                        upload_files = files.UploadFiles.get(upload_id, is_authorized=lambda: True)
                        upload_files_cache[upload_id] = upload_files

                    try:
                        entry_metadata = calc.entry_metadata(upload_files)
                        entry = entry_metadata.a_elastic.create_index_entry().to_dict(include_meta=True)
                        entry['_op_type'] = 'index'

                        yield entry

                    except Exception as e:
                        common.logger.error('edit repo could not create index doc', exc_info=e)

                for upload_files in upload_files_cache.values():
                    upload_files.close()

            _, failed = elasticsearch.helpers.bulk(
                infrastructure.elastic_client, elastic_updates(), stats_only=True)
            search.refresh()
            if failed > 0:
                common.logger.error(
                    'edit repo with failed elastic updates',
                    payload=mongo_update, nfailed=len(failed))

    return list(upload_ids)


def get_uploader_ids(query):
    ''' Get all the uploader from the query, to check coauthers and shared_with for uploaders. '''
    search_request = search.SearchRequest()
    apply_search_parameters(search_request, query)
    search_request.quantity(name='uploader_id')
    return search_request.execute()['quantities']['uploader_id']['values']


@ns.route('/edit')
class EditRepoCalcsResource(Resource):
    @api.doc('edit_repo')
    @api.response(400, 'Invalid requests, e.g. wrong owner type or bad search parameters')
    @api.expect(_repo_edit_model)
    @api.marshal_with(_repo_edit_model, skip_none=True, code=200, description='Edit verified/performed')
    @authenticate()
    def post(self):
        ''' Edit repository metadata. '''

        # basic body parsing and some semantic checks
        json_data = request.get_json()
        if json_data is None:
            json_data = {}
        query = json_data.get('query', {})

        owner = query.get('owner', 'user')
        if owner not in ['user', 'staging']:
            abort(400, 'Not a valid owner for edit %s. Edit can only be performed in user or staging' % owner)
        query['owner'] = owner

        if 'actions' not in json_data:
            abort(400, 'Missing key actions in edit data')
        actions = json_data['actions']
        verify = json_data.get('verify', False)

        # preparing the query of entries that are edited
        parsed_query = {}
        for quantity_name, value in query.items():
            if quantity_name in _search_quantities:
                quantity = search.search_quantities[quantity_name]
                if quantity.many:
                    if not isinstance(value, list):
                        value = value.split(',')
                parsed_query[quantity_name] = value
        parsed_query['owner'] = owner
        parsed_query['domain'] = query.get('domain')

        # checking the edit actions and preparing a mongo update on the fly
        json_data['success'] = True
        mongo_update = {}
        uploader_ids = None
        lift_embargo = False
        has_error = False
        removed_datasets = None

        with utils.timer(common.logger, 'edit verified'):
            for action_quantity_name, quantity_actions in actions.items():
                quantity = _editable_quantities.get(action_quantity_name)
                if quantity is None:
                    abort(400, 'Unknown quantity %s' % action_quantity_name)

                quantity_flask = quantity.m_get_annotations('flask', {})
                if quantity_flask.get('admin_only', False):
                    if not g.user.is_admin():
                        abort(404, 'Only the admin user can set %s' % quantity.name)

                if isinstance(quantity_actions, list) == quantity.is_scalar:
                    abort(400, 'Wrong shape for quantity %s' % action_quantity_name)

                if not isinstance(quantity_actions, list):
                    quantity_actions = [quantity_actions]

                flask_verify = quantity_flask.get('verify', None)
                mongo_key = 'metadata__%s' % quantity.name
                has_error = False
                for action in quantity_actions:
                    action['success'] = True
                    action['message'] = None
                    action_value = action.get('value')
                    action_value = action_value if action_value is None else action_value.strip()

                    if action_quantity_name == 'with_embargo':
                        # ignore the actual value ... just lift the embargo
                        mongo_value = False
                        lift_embargo = True

                        # check if necessary
                        search_request = search.SearchRequest()
                        apply_search_parameters(search_request, parsed_query)
                        search_request.q = search_request.q & Q('term', with_embargo=True)
                        if search_request.execute()['total'] == 0:
                            action['success'] = False
                            has_error = True
                            action['message'] = 'There is no embargo to lift'
                            continue

                    elif action_value is None:
                        mongo_value = None

                    elif action_value == '':
                        mongo_value = None

                    elif flask_verify == datamodel.User:
                        try:
                            mongo_value = User.get(user_id=action_value).user_id
                        except KeyError:
                            action['success'] = False
                            has_error = True
                            action['message'] = 'User does not exist'
                            continue

                        if uploader_ids is None:
                            uploader_ids = get_uploader_ids(parsed_query)
                        if action_value in uploader_ids:
                            action['success'] = False
                            has_error = True
                            action['message'] = 'This user is already an uploader of one entry in the query'
                            continue

                    elif flask_verify == datamodel.Dataset:
                        try:
                            mongo_value = Dataset.m_def.a_mongo.get(
                                user_id=g.user.user_id, name=action_value).dataset_id
                        except KeyError:
                            action['message'] = 'Dataset does not exist and will be created'
                            mongo_value = None
                            if not verify:
                                dataset = Dataset(
                                    dataset_id=utils.create_uuid(), user_id=g.user.user_id,
                                    name=action_value, created=datetime.utcnow())
                                dataset.a_mongo.create()
                                mongo_value = dataset.dataset_id

                    else:
                        mongo_value = action_value

                    # verify if the edit action creates consistent shared_with and with_embargo for
                    # whole uploads
                    if action_quantity_name == 'shared_with' or action_quantity_name == 'with_embargo':
                        search_request = search.SearchRequest()
                        apply_search_parameters(search_request, parsed_query)
                        search_request.quantity('upload_id')
                        uploads = search_request.execute()['quantities']['upload_id']['values']
                        for upload_id, upload_data in uploads.items():
                            search_request = search.SearchRequest().search_parameters(upload_id=upload_id)
                            entries = search_request.execute()['total']
                            if entries != upload_data['total']:
                                action['success'] = False
                                action['message'] = json_data.get('message', '') + (
                                    'Edit would create an upload with inconsistent shared_with or with_embargo. '
                                    'You can only set those for all entries of an upload.')
                                has_error = True
                                continue

                    if len(quantity.shape) == 0:
                        mongo_update[mongo_key] = mongo_value
                    else:
                        mongo_values = mongo_update.setdefault(mongo_key, [])
                        if mongo_value is not None:
                            if mongo_value in mongo_values:
                                action['success'] = False
                                has_error = True
                                action['message'] = 'Duplicate values are not allowed'
                                continue
                            mongo_values.append(mongo_value)

                if len(quantity_actions) == 0 and len(quantity.shape) > 0:
                    mongo_update[mongo_key] = []

                if action_quantity_name == 'datasets':
                    # check if datasets edit is allowed and if datasets have to be removed
                    search_request = search.SearchRequest()
                    apply_search_parameters(search_request, parsed_query)
                    search_request.quantity(name='dataset_id')
                    old_datasets = list(
                        search_request.execute()['quantities']['dataset_id']['values'].keys())

                    removed_datasets = []
                    for dataset_id in old_datasets:
                        if dataset_id not in mongo_update.get(mongo_key, []):
                            removed_datasets.append(dataset_id)

                    doi_ds = Dataset.m_def.a_mongo.objects(
                        dataset_id__in=removed_datasets, doi__ne=None).first()
                    if doi_ds is not None:
                        json_data['success'] = False
                        json_data['message'] = json_data.get('message', '') + \
                            'Edit would remove entries from a dataset with DOI (%s) ' % doi_ds.name
                        has_error = True

        # stop here, if client just wants to verify its actions
        if verify:
            return json_data, 200

        # stop if the action were not ok
        if has_error:
            return json_data, 400

        # perform the change
        mongo_update['metadata__last_edit'] = datetime.utcnow()
        upload_ids = edit(parsed_query, mongo_update, True)

        # lift embargo
        if lift_embargo:
            for upload_id in upload_ids:
                upload = proc.Upload.get(upload_id)
                if upload.published:
                    upload.re_pack()

        # remove potentially empty old datasets
        if removed_datasets is not None:
            for dataset in removed_datasets:
                if proc.Calc.objects(metadata__datasets=dataset).first() is None:
                    Dataset.m_def.a_mongo.objects(dataset_id=dataset).delete()

        return json_data, 200


_repo_quantity_search_request_parser = api.parser()
add_search_parameters(_repo_quantity_search_request_parser)
_repo_quantity_search_request_parser.add_argument(
    'after', type=str, help='The after value to use for "scrolling".')
_repo_quantity_search_request_parser.add_argument(
    'size', type=int, help='The max size of the returned values.')
_repo_quantity_search_request_parser.add_argument(
    'value', type=str, help='A partial value. Only values that include this will be returned')

_repo_quantity_model = api.model('RepoQuantity', {
    'after': fields.String(description='The after value that can be used to retrieve the next set of values.'),
    'values': fields.Raw(description='A dict with values as key. Values are dicts with "total" and "examples" keys.')
})

_repo_quantity_values_model = api.model('RepoQuantityValues', {
    'quantity': fields.Nested(_repo_quantity_model, allow_null=True)
})


@ns.route('/quantity/<string:quantity>')
class RepoQuantityResource(Resource):
    @api.doc('quantity_search')
    @api.response(400, 'Invalid requests, e.g. wrong owner type, bad quantity, bad search parameters')
    @api.expect(_repo_quantity_search_request_parser, validate=True)
    @api.marshal_with(_repo_quantity_values_model, skip_none=True, code=200, description='Search results send')
    @authenticate()
    def get(self, quantity: str):
        '''
        Retrieve quantity values from entries matching the search.

        You can use the various quantities to search/filter for. For some of the
        indexed quantities this endpoint returns aggregation information. This means
        you will be given a list of all possible values and the number of entries
        that have the certain value. You can also use these aggregations on an empty
        search to determine the possible values.

        There is no ordering and no pagination. Instead there is an 'after' key based
        scrolling. The result will contain an 'after' value, that can be specified
        for the next request. You can use the 'size' and 'after' parameters accordingly.

        The result will contain a 'quantity' key with quantity values and the "after"
        value. There will be upto 'size' many values. For the rest of the values use the
        "after" parameter in another request.
        '''

        search_request = search.SearchRequest()
        args = {
            key: value
            for key, value in _repo_quantity_search_request_parser.parse_args().items()
            if value is not None}

        apply_search_parameters(search_request, args)
        after = args.get('after', None)
        size = args.get('size', 100)

        try:
            assert size >= 0
        except AssertionError:
            abort(400, message='invalid size')

        try:
            search_request.quantity(quantity, size=size, after=after)
            results = search_request.execute()
            quantities = results.pop('quantities')
            results['quantity'] = quantities[quantity]

            return results, 200
        except KeyError as e:
            import traceback
            traceback.print_exc()
            abort(400, 'Given quantity does not exist: %s' % str(e))


_repo_suggestions_search_request_parser = api.parser()
add_search_parameters(_repo_suggestions_search_request_parser)
_repo_suggestions_search_request_parser.add_argument(
    'size', type=int, help='The max size of the returned values.')
_repo_suggestions_search_request_parser.add_argument(
    'include', type=str, help='A substring that all values need to include.')

_repo_suggestions_model = api.model('RepoSuggestionsValues', {
    'suggestions': fields.List(fields.String, description='A list with the suggested values.')
})


@ns.route('/suggestions/<string:quantity>')
class RepoSuggestionsResource(Resource):
    @api.doc('suggestions_search')
    @api.response(400, 'Invalid requests, e.g. wrong owner type, bad quantity, bad search parameters')
    @api.expect(_repo_suggestions_search_request_parser, validate=True)
    @api.marshal_with(_repo_suggestions_model, skip_none=True, code=200, description='Suggestions send')
    @authenticate()
    def get(self, quantity: str):
        '''
        Retrieve the top values for the given quantity from entries matching the search.
        Values can be filtered by to include a given value.

        There is no ordering, no pagination, and no scroll interface.

        The result will contain a 'suggestions' key with values. There will be upto 'size' many values.
        '''

        search_request = search.SearchRequest()
        args = {
            key: value
            for key, value in _repo_suggestions_search_request_parser.parse_args().items()
            if value is not None}

        apply_search_parameters(search_request, args)
        size = args.get('size', 20)
        include = args.get('include', None)

        try:
            assert size >= 0
        except AssertionError:
            abort(400, message='invalid size')

        try:
            search_request.statistic(quantity, size=size, include=include, order=dict(_key='desc'))
            results = search_request.execute()
            values = {
                value: metric['code_runs']
                for value, metric in results['statistics'][quantity].items()
                if metric['code_runs'] > 0}
            results['suggestions'] = sorted(
                values.keys(), key=lambda value: values[value], reverse=True)

            return results, 200
        except KeyError as e:
            import traceback
            traceback.print_exc()
            abort(400, 'Given quantity does not exist: %s' % str(e))


_repo_quantities_search_request_parser = api.parser()
add_search_parameters(_repo_quantities_search_request_parser)
_repo_quantities_search_request_parser.add_argument(
    'quantities', type=str, action='append',
    help='The quantities to retrieve values from')
_repo_quantities_search_request_parser.add_argument(
    'size', type=int, help='The max size of the returned values.')

_repo_quantities_model = api.model('RepoQuantitiesResponse', {
    'quantities': fields.Nested(api.model('RepoQuantities', {
        quantity: fields.List(fields.Nested(_repo_quantity_model))
        for quantity in search.search_quantities
    }))
})


@ns.route('/quantities')
class RepoQuantitiesResource(Resource):
    @api.doc('quantities_search')
    @api.response(400, 'Invalid requests, e.g. wrong owner type, bad quantity, bad search parameters')
    @api.expect(_repo_quantities_search_request_parser, validate=True)
    @api.marshal_with(_repo_quantities_model, skip_none=True, code=200, description='Search results send')
    @authenticate()
    def get(self):
        '''
        Retrieve quantity values for multiple quantities at once.

        You can use the various quantities to search/filter for. For some of the
        indexed quantities this endpoint returns aggregation information. This means
        you will be given a list of all possible values and the number of entries
        that have the certain value. You can also use these aggregations on an empty
        search to determine the possible values.

        There is no ordering and no pagination and not after key based scrolling. Instead
        there is an 'after' key based scrolling.

        The result will contain a 'quantities' key with a dict of quantity names and the
        retrieved values as values.
        '''

        search_request = search.SearchRequest()
        args = {
            key: value
            for key, value in _repo_quantities_search_request_parser.parse_args().items()
            if value is not None}

        apply_search_parameters(search_request, args)
        quantities = args.get('quantities', [])
        size = args.get('size', 5)

        try:
            assert size >= 0
        except AssertionError:
            abort(400, message='invalid size')

        for quantity in quantities:
            try:
                search_request.quantity(quantity, size=size)
            except KeyError as e:
                import traceback
                traceback.print_exc()
                abort(400, 'Given quantity does not exist: %s' % str(e))

        return search_request.execute(), 200


_repo_calc_id_model = api.model('RepoCalculationId', {
    'upload_id': fields.String(), 'calc_id': fields.String()
})


@ns.route('/pid/<path:pid>')
class RepoPidResource(Resource):
    @api.doc('resolve_pid')
    @api.response(404, 'Entry with PID does not exist')
    @api.marshal_with(_repo_calc_id_model, skip_none=True, code=200, description='Entry resolved')
    @authenticate()
    def get(self, pid: str):
        if '/' in pid:
            prefix, pid = pid.split('/')
            if prefix != '21.11132':
                abort(400, 'Wrong PID format')
            try:
                pid_int = utils.decode_handle_id(pid)
            except ValueError:
                abort(400, 'Wrong PID format')
        else:
            try:
                pid_int = int(pid)
            except ValueError:
                abort(400, 'Wrong PID format')

        search_request = search.SearchRequest().include('upload_id', 'calc_id')

        if g.user is not None:
            search_request.owner('all', user_id=g.user.user_id)
        else:
            search_request.owner('all')

        search_request.search_parameter('pid', pid_int)

        results = list(search_request.execute_scan())
        total = len(results)

        if total == 0:
            abort(404, 'Entry with PID %s does not exist' % pid)

        if total > 1:
            common.logger.error('Two entries for the same pid', pid=pid_int)

        result = results[0]
        return dict(
            upload_id=result['upload_id'],
            calc_id=result['calc_id'])



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.cli import lazy_import

lazy_import.lazy_module('typing')
lazy_import.lazy_module('click')
lazy_import.lazy_module('asyncio')
lazy_import.lazy_module('concurrent.futures')
lazy_import.lazy_module('datetime')
lazy_import.lazy_module('time')
lazy_import.lazy_module('elasticsearch_dsl')
lazy_import.lazy_module('elasticsearch')
lazy_import.lazy_module('os')
lazy_import.lazy_module('shutil')
lazy_import.lazy_module('tabulate')
lazy_import.lazy_module('sys')
lazy_import.lazy_module('io')
lazy_import.lazy_module('re')
lazy_import.lazy_module('uuid')
lazy_import.lazy_module('json')
lazy_import.lazy_module('threading')
lazy_import.lazy_module('numpy')
lazy_import.lazy_module('requests')
lazy_import.lazy_module('pymongo')
lazy_import.lazy_module('mongoengine')
lazy_import.lazy_module('ase')
lazy_import.lazy_module('bs4')
lazy_import.lazy_module('matid')
lazy_import.lazy_module('matid.symmetry.symmetryanalyzer')
lazy_import.lazy_module('matid.utils.segfault_protect')
lazy_import.lazy_module('nomad.atomutils')
lazy_import.lazy_module('nomad.normalizing')
lazy_import.lazy_module('nomad.processing')
lazy_import.lazy_module('nomad.search')
lazy_import.lazy_module('nomad.datamodel')
lazy_import.lazy_module('nomad.infrastructure')
lazy_import.lazy_module('nomad.utils')
lazy_import.lazy_module('nomad.config')
lazy_import.lazy_module('nomad.files')
lazy_import.lazy_module('nomad.archive')

from . import admin, uploads, entries, run, clean, users, migrate  # noqa



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='gaussian_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='gaussian.nomadmetainfo.json'))


class x_gaussian_configuration_core(MCategory):
    '''
    Properties defining the current configuration.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_gaussian_configuration_core'))


class x_gaussian_atom_forces_type(MCategory):
    '''
    Some forces on the atoms (i.e. minus derivatives of some energy with respect to the
    atom position).
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_gaussian_atom_forces_type'))


class x_gaussian_scf_info(MCategory):
    '''
    Information on the self-consistent field (SCF) procedure.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_gaussian_scf_info'))


class x_gaussian_section_geometry(MSection):
    '''
    section that contains Cartesian coordinates of the system for a given geometry
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_geometry'))

    x_gaussian_atomic_number = Quantity(
        type=str,
        shape=[],
        description='''
        atomic number for atoms
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atomic_number'))

    x_gaussian_atom_x_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x coordinate for the atoms
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_x_coord'))

    x_gaussian_atom_y_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y coordinate for the atoms
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_y_coord'))

    x_gaussian_atom_z_coord = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z coordinate for the atoms
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_z_coord'))


class x_gaussian_section_hybrid_coeffs(MSection):
    '''
    section that contains coefficients for the hybrid DFT functionals
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_hybrid_coeffs'))

    hybrid_xc_coeff1 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Coefficient for Hartree-Fock exchange in hybrid DFT functionals
        ''',
        a_legacy=LegacyDefinition(name='hybrid_xc_coeff1'))

    hybrid_xc_coeff2 = Quantity(
        type=str,
        shape=[],
        description='''
        Coefficients for Slater exchange, non-local exchange, local correlation, and non-
        local correlation, respectively, in hybrid DFT functionals
        ''',
        a_legacy=LegacyDefinition(name='hybrid_xc_coeff2'))

    x_gaussian_hybrid_xc_hfx = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hartree-Fock exchange
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hybrid_xc_hfx'))

    x_gaussian_hybrid_xc_slater = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Slater exchange
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hybrid_xc_slater'))

    x_gaussian_hybrid_xc_nonlocalex = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Slater exchange
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hybrid_xc_nonlocalex'))

    x_gaussian_hybrid_xc_localcorr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Slater exchange
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hybrid_xc_localcorr'))

    x_gaussian_hybrid_xc_nonlocalcorr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Slater exchange
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hybrid_xc_nonlocalcorr'))


class x_gaussian_section_total_scf_one_geometry(MSection):
    '''
    Check for SCF convergence and writes the total energy value to backend
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_total_scf_one_geometry'))


class x_gaussian_section_times(MSection):
    '''
    section that contains the execution times of the run
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_times'))

    x_gaussian_program_termination_date = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_program_termination_date'))

    x_gaussian_program_cpu_time = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_program_cpu_time'))


class x_gaussian_section_atom_forces(MSection):
    '''
    section that contains Cartesian coordinates of the system for a given geometry
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_atom_forces'))

    x_gaussian_atom_x_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_x_force'))

    x_gaussian_atom_y_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_y_force'))

    x_gaussian_atom_z_force = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atom_z_force'))


class x_gaussian_section_molecular_multipoles(MSection):
    '''
    Section describing multipoles (charges, dipoles,...).
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_molecular_multipoles'))

    charge = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the total charge of the system (in electronic units).
        ''',
        a_legacy=LegacyDefinition(name='charge'))

    dipole_moment_x = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the x component of the dipole moment (Debye).
        ''',
        a_legacy=LegacyDefinition(name='dipole_moment_x'))

    dipole_moment_y = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the y component of the dipole moment (Debye).
        ''',
        a_legacy=LegacyDefinition(name='dipole_moment_y'))

    dipole_moment_z = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the z component of the dipole moment (Debye).
        ''',
        a_legacy=LegacyDefinition(name='dipole_moment_z'))

    quadrupole_moment_xx = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xx component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_xx'))

    quadrupole_moment_yy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yy component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_yy'))

    quadrupole_moment_zz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zz component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_zz'))

    quadrupole_moment_xy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xy component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_xy'))

    quadrupole_moment_xz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xz component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_xz'))

    quadrupole_moment_yz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yz component of the quadrupole moment (Debye-Ang).
        ''',
        a_legacy=LegacyDefinition(name='quadrupole_moment_yz'))

    octapole_moment_xxx = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxx component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xxx'))

    octapole_moment_yyy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyy component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_yyy'))

    octapole_moment_zzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zzz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_zzz'))

    octapole_moment_xyy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xyy component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xyy'))

    octapole_moment_xxy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxy component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xxy'))

    octapole_moment_xxz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xxz'))

    octapole_moment_xzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xzz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xzz'))

    octapole_moment_yzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yzz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_yzz'))

    octapole_moment_yyz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_yyz'))

    octapole_moment_xyz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xyz component of the octapole moment (Debye-Ang**2).
        ''',
        a_legacy=LegacyDefinition(name='octapole_moment_xyz'))

    hexadecapole_moment_xxxx = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxxx component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxxx'))

    hexadecapole_moment_yyyy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyyy component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_yyyy'))

    hexadecapole_moment_zzzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zzzz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_zzzz'))

    hexadecapole_moment_xxxy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxxy component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxxy'))

    hexadecapole_moment_xxxz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxxz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxxz'))

    hexadecapole_moment_yyyx = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyyx component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_yyyx'))

    hexadecapole_moment_yyyz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyyz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_yyyz'))

    hexadecapole_moment_zzzx = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zzzx component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_zzzx'))

    hexadecapole_moment_zzzy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zzzy component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_zzzy'))

    hexadecapole_moment_xxyy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxyy component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxyy'))

    hexadecapole_moment_xxzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxzz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxzz'))

    hexadecapole_moment_yyzz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyzz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_yyzz'))

    hexadecapole_moment_xxyz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the xxyz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_xxyz'))

    hexadecapole_moment_yyxz = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the yyxz component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_yyxz'))

    hexadecapole_moment_zzxy = Quantity(
        type=str,
        shape=[],
        description='''
        Value of the zzxy component of the hexadecapole moment (Debye-Ang**3).
        ''',
        a_legacy=LegacyDefinition(name='hexadecapole_moment_zzxy'))

    x_gaussian_molecular_multipole_lm = Quantity(
        type=np.dtype(np.int32),
        shape=['x_gaussian_number_of_lm_molecular_multipoles', 2],
        description='''
        Tuples of $l$ and $m$ values for which the molecular multipoles (including the
        electric charge, dipole, etc.) are given. The meaning of the integer number $l$ is
        monopole/charge for $l=0$, dipole for $l=1$, quadrupole for $l=2$, etc. The
        meaning of the integer numbers $m$ is specified by molecular_multipole_m_kind.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_molecular_multipole_lm'))

    x_gaussian_molecular_multipole_m_kind = Quantity(
        type=str,
        shape=[],
        description='''
        String describing what the integer numbers $m$ in molecular_multipole_lm mean.
        Allowed values (for atomic multipoles) are listed in the [m\\_kind wiki
        page](https://gitlab.rzg.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/m-kind).
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_molecular_multipole_m_kind'))

    x_gaussian_molecular_multipole_values = Quantity(
        type=np.dtype(np.float64),
        shape=['x_gaussian_number_of_lm_molecular_multipoles'],
        description='''
        Value of the multipoles (including the monopole/charge for $l$ = 0, the dipole for
        $l$ = 1, etc.).
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_molecular_multipole_values'))

    x_gaussian_number_of_lm_molecular_multipoles = Quantity(
        type=int,
        shape=[],
        description='''
        Number of $l,m$ combinations for which molecular multipoles are given.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_number_of_lm_molecular_multipoles'))


class x_gaussian_section_geometry_optimization_info(MSection):
    '''
    Specifies whether a geometry optimization is converged.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_geometry_optimization_info'))

    x_gaussian_geometry_optimization_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Specifies whether a geometry optimization is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_geometry_optimization_converged'))


class x_gaussian_section_frequencies(MSection):
    '''
    section for the values of the frequencies, reduced masses and normal mode vectors
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_frequencies'))

    x_gaussian_frequency_values = Quantity(
        type=str,
        shape=['number_of_frequency_rows'],
        description='''
        values of frequencies, in cm-1
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_frequency_values'))

    x_gaussian_frequencies = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_frequencies'],
        description='''
        values of frequencies, in cm-1
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_frequencies'))

    x_gaussian_reduced_masses = Quantity(
        type=str,
        shape=['number_of_reduced_masses_rows'],
        description='''
        values of normal mode reduced masses
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_reduced_masses'))

    x_gaussian_red_masses = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_frequencies'],
        description='''
        values of normal mode reduced masses
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_red_masses'))

    x_gaussian_normal_modes = Quantity(
        type=str,
        shape=['number_of_normal_modes_rows'],
        description='''
        normal mode vectors
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_normal_modes'))

    x_gaussian_normal_mode_values = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_frequencies', 'number_of_atoms', 3],
        description='''
        normal mode vectors
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_normal_mode_values'))


class x_gaussian_section_thermochem(MSection):
    '''
    section for thermochemical quantities
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_thermochem'))

    x_gaussian_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of temperature for thermochemical values
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_temperature'))

    x_gaussian_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of pressure for thermochemical values
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_pressure'))

    x_gaussian_moment_of_inertia_X = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        X component of moment of inertia
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_moment_of_inertia_X'))

    x_gaussian_moment_of_inertia_Y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Y component of moment of inertia
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_moment_of_inertia_Y'))

    x_gaussian_moment_of_inertia_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Z component of moment of inertia
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_moment_of_inertia_Z'))

    x_gaussian_moments = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Values of moments of inertia
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_moments'))

    x_gaussian_zero_point_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of zero-point energy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_zero_point_energy'))

    x_gaussian_thermal_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of thermal correction to total energy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_thermal_correction_energy'))

    x_gaussian_thermal_correction_enthalpy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of thermal correction to enthalpy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_thermal_correction_enthalpy'))

    x_gaussian_thermal_correction_free_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Value of thermal correction to Gibbs free energy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_thermal_correction_free_energy'))


class x_gaussian_section_force_constant_matrix(MSection):
    '''
    section for force constant matrix in Cartesians. Units are mdyne.Angstrom
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_force_constant_matrix'))

    x_gaussian_force_constants = Quantity(
        type=str,
        shape=[],
        description='''
        Force constant matrix elements
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_force_constants'))

    x_gaussian_force_constant_values = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_degrees_of_freedom', 'number_of_degrees_of_freedom'],
        description='''
        Force constant matrix element values
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_force_constant_values'))


class x_gaussian_section_orbital_symmetries(MSection):
    '''
    section for the symmetry of the MOs
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_orbital_symmetries'))

    x_gaussian_alpha_occ_symmetry_values = Quantity(
        type=str,
        shape=['number_of_alpha_occ_rows'],
        description='''
        symmetry of the alpha occupied MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_occ_symmetry_values'))

    x_gaussian_alpha_vir_symmetry_values = Quantity(
        type=str,
        shape=['number_of_alpha_vir_rows'],
        description='''
        symmetry of the alpha virtual MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_vir_symmetry_values'))

    x_gaussian_beta_occ_symmetry_values = Quantity(
        type=str,
        shape=['number_of_beta_occ_rows'],
        description='''
        symmetry of the beta occupied MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_occ_symmetry_values'))

    x_gaussian_beta_vir_symmetry_values = Quantity(
        type=str,
        shape=['number_of_beta_vir_rows'],
        description='''
        symmetry of the beta virtual MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_vir_symmetry_values'))

    x_gaussian_alpha_symmetries = Quantity(
        type=str,
        shape=['number_of_alpha_mos'],
        description='''
        symmetry of the alpha MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_symmetries'))

    x_gaussian_beta_symmetries = Quantity(
        type=str,
        shape=['number_of_beta_mos'],
        description='''
        symmetry of the beta MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_symmetries'))


class x_gaussian_section_symmetry(MSection):
    '''
    section for the symmetry of the electronic state
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_symmetry'))

    x_gaussian_elstate_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        symmetry group of the electronic state
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_elstate_symmetry'))


class x_gaussian_section_elstruc_method(MSection):
    '''
    Section containing electronic structure method.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_elstruc_method'))

    x_gaussian_electronic_structure_method = Quantity(
        type=str,
        shape=[],
        description='''
        Name of electronic structure method.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_electronic_structure_method'))


class x_gaussian_section_energy_components(MSection):
    '''
    Section containing total energy components
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_energy_components'))


class x_gaussian_section_moller_plesset(MSection):
    '''
    Perturbative Moller-Plesset energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_moller_plesset'))

    x_gaussian_mp2_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP2 energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp2_correction_energy'))

    x_gaussian_mp3_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP3 energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp3_correction_energy'))

    x_gaussian_mp4dq_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP4(DQ) energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp4dq_correction_energy'))

    x_gaussian_mp4sdq_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP4(SDQ) energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp4sdq_correction_energy'))

    x_gaussian_mp4sdtq_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP4(SDTQ) energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp4sdtq_correction_energy'))

    x_gaussian_mp5_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and MP5 energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_mp5_correction_energy'))


class x_gaussian_section_coupled_cluster(MSection):
    '''
    Coupled cluster energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_coupled_cluster'))

    x_gaussian_ccsd_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and CCSD energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_ccsd_correction_energy'))


class x_gaussian_section_quadratic_ci(MSection):
    '''
    Quadratic CI energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_quadratic_ci'))

    x_gaussian_qcisd_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and QCISD energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_qcisd_correction_energy'))

    x_gaussian_qcisdtq_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and QCISD(TQ) energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_qcisdtq_correction_energy'))


class x_gaussian_section_ci(MSection):
    '''
    CI energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_ci'))

    x_gaussian_ci_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Difference between SCF and CI energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_ci_correction_energy'))


class x_gaussian_section_semiempirical(MSection):
    '''
    semiempirical convergence cycles and energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_semiempirical'))

    x_gaussian_semiempirical_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        semiempirical energies.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_semiempirical_energy'))

    x_gaussian_semiempirical_method = Quantity(
        type=str,
        shape=[],
        description='''
        semiempirical method.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_semiempirical_method'))


class x_gaussian_section_molmech(MSection):
    '''
    molecular mechanics method and energies.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_molmech'))

    x_gaussian_molmech_method = Quantity(
        type=str,
        shape=[],
        description='''
        molecular mechanics method.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_molmech_method'))


class x_gaussian_section_models(MSection):
    '''
    composite model chemistries.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_models'))


class x_gaussian_section_excited_initial(MSection):
    '''
    Excited state energies and properties
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_excited_initial'))


class x_gaussian_section_excited(MSection):
    '''
    CI singles, TDDFT/TDHF, ZINDO or EOMCCSD excited state energies and properties
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_excited'))

    x_gaussian_excited_state_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CIS, TDDFT/TDHF, ZINDO or EOMCCSD excited state number
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_excited_state_number'))

    x_gaussian_excited_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        CIS, TDDFT/TDHF, ZINDO or EOMCCSD excited state energy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_excited_energy'))

    x_gaussian_excited_oscstrength = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CIS, TDDFT/TDHF, ZINDO or EOMCCSD excited state oscillator strength
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_excited_oscstrength'))

    x_gaussian_excited_spin_squared = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CIS, TDDFT/TDHF, ZINDO or EOMCCSD excited state spin squared value
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_excited_spin_squared'))

    x_gaussian_excited_transition = Quantity(
        type=str,
        shape=[],
        description='''
        CIS, TDDFT/TDHF, ZINDO or EOMCCSD excited state MOs involved in transitions and
        their coefficients
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_excited_transition'))


class x_gaussian_section_casscf(MSection):
    '''
    CASSCF energies and properties
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_gaussian_section_casscf'))

    x_gaussian_casscf_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        CASSCF energy
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_casscf_energy'))

    x_gaussian_casscf_determinant = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        CASSCF determinant number
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_casscf_determinant'))

    x_gaussian_casscf_coefficient = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        CASSCF determinant coefficient
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_casscf_coefficient'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_gaussian_number_of_atoms = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of atoms of the system
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_number_of_atoms'))

    x_gaussian_atomic_masses = Quantity(
        type=str,
        shape=[],
        description='''
        atomic masses for atoms
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_atomic_masses'))

    x_gaussian_total_charge = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Total charge of the system.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_total_charge'))

    x_gaussian_spin_target_multiplicity = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Target (user-imposed) value of the spin multiplicity $M=2S+1$, where $S$ is the
        total spin. It is an integer value. This value is not necessarly the value
        obtained at the end of the calculation. See spin_S2 for the converged value of the
        spin moment.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_spin_target_multiplicity'))

    x_gaussian_geometry_lattice_vector_x = Quantity(
        type=str,
        shape=[],
        description='''
        x component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_geometry_lattice_vector_x'))

    x_gaussian_geometry_lattice_vector_y = Quantity(
        type=str,
        shape=[],
        description='''
        y component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_geometry_lattice_vector_y'))

    x_gaussian_geometry_lattice_vector_z = Quantity(
        type=str,
        shape=[],
        description='''
        z component of lattice vector
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_geometry_lattice_vector_z'))

    x_gaussian_masses = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        values of atomic masses
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_masses'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_gaussian_atom_forces = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='newton',
        description='''
        Forces acting on the atoms, calculated as minus gradient of energy_total.
        ''',
        categories=[x_gaussian_atom_forces_type],
        a_legacy=LegacyDefinition(name='x_gaussian_atom_forces'))

    x_gaussian_energy_total_scf_converged = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        SCF total energy converged for a given geometry.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_energy_total_scf_converged'))

    x_gaussian_number_of_scf_iterations = Quantity(
        type=int,
        shape=[],
        description='''
        Number of performed self-consistent field (SCF) iterations at DFT level.
        ''',
        categories=[x_gaussian_scf_info],
        a_legacy=LegacyDefinition(name='x_gaussian_number_of_scf_iterations'))

    x_gaussian_section_hybrid_coeffs = SubSection(
        sub_section=SectionProxy('x_gaussian_section_hybrid_coeffs'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_hybrid_coeffs'))

    x_gaussian_section_total_scf_one_geometry = SubSection(
        sub_section=SectionProxy('x_gaussian_section_total_scf_one_geometry'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_total_scf_one_geometry'))

    x_gaussian_section_atom_forces = SubSection(
        sub_section=SectionProxy('x_gaussian_section_atom_forces'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_atom_forces'))

    x_gaussian_section_molecular_multipoles = SubSection(
        sub_section=SectionProxy('x_gaussian_section_molecular_multipoles'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_molecular_multipoles'))

    x_gaussian_section_energy_components = SubSection(
        sub_section=SectionProxy('x_gaussian_section_energy_components'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_energy_components'))

    x_gaussian_section_moller_plesset = SubSection(
        sub_section=SectionProxy('x_gaussian_section_moller_plesset'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_moller_plesset'))

    x_gaussian_section_coupled_cluster = SubSection(
        sub_section=SectionProxy('x_gaussian_section_coupled_cluster'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_coupled_cluster'))

    x_gaussian_section_quadratic_ci = SubSection(
        sub_section=SectionProxy('x_gaussian_section_quadratic_ci'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_quadratic_ci'))

    x_gaussian_section_ci = SubSection(
        sub_section=SectionProxy('x_gaussian_section_ci'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_ci'))

    x_gaussian_section_semiempirical = SubSection(
        sub_section=SectionProxy('x_gaussian_section_semiempirical'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_semiempirical'))

    x_gaussian_section_molmech = SubSection(
        sub_section=SectionProxy('x_gaussian_section_molmech'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_molmech'))

    x_gaussian_section_models = SubSection(
        sub_section=SectionProxy('x_gaussian_section_models'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_models'))

    x_gaussian_section_excited_initial = SubSection(
        sub_section=SectionProxy('x_gaussian_section_excited_initial'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_excited_initial'))

    x_gaussian_section_excited = SubSection(
        sub_section=SectionProxy('x_gaussian_section_excited'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_excited'))

    x_gaussian_section_casscf = SubSection(
        sub_section=SectionProxy('x_gaussian_section_casscf'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_casscf'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_gaussian_chk_file = Quantity(
        type=str,
        shape=[],
        description='''
        binary file with detailed output information
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_chk_file'))

    x_gaussian_memory = Quantity(
        type=str,
        shape=[],
        description='''
        total memory for the run
        ''',
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_gaussian_memory'))

    x_gaussian_number_of_processors = Quantity(
        type=str,
        shape=[],
        description='''
        number of processors used
        ''',
        categories=[public.settings_run, public.parallelization_info, public.accessory_info],
        a_legacy=LegacyDefinition(name='x_gaussian_number_of_processors'))

    x_gaussian_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_program_execution_date'))

    x_gaussian_program_release_date = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_program_release_date'))

    x_gaussian_program_implementation = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_program_implementation'))

    x_gaussian_atom_positions = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Positions of all the atoms, in Cartesian coordinates. This metadata defines a
        configuration and is therefore required.
        ''',
        categories=[x_gaussian_configuration_core],
        a_legacy=LegacyDefinition(name='x_gaussian_atom_positions'))

    x_gaussian_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        Labels of the atoms. These strings identify the atom kind and conventionally start
        with the symbol of the atomic species, possibly followed by a number. The same
        atomic species can be labelled with more than one atom_labels in order to
        distinguish, e.g., atoms of the same species assigned to different atom-centered
        basis sets or pseudopotentials, or simply atoms in different locations in the
        structure (e.g., bulk and surface). These labels can also be used for *particles*
        that do not correspond to physical atoms (e.g., ghost atoms in some codes using
        atom-centered basis sets). This metadata defines a configuration and is therefore
        required.
        ''',
        categories=[x_gaussian_configuration_core],
        a_legacy=LegacyDefinition(name='x_gaussian_atom_labels'))

    x_gaussian_section_geometry = SubSection(
        sub_section=SectionProxy('x_gaussian_section_geometry'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_geometry'))

    x_gaussian_section_times = SubSection(
        sub_section=SectionProxy('x_gaussian_section_times'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_times'))

    x_gaussian_section_geometry_optimization_info = SubSection(
        sub_section=SectionProxy('x_gaussian_section_geometry_optimization_info'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_geometry_optimization_info'))

    x_gaussian_section_frequencies = SubSection(
        sub_section=SectionProxy('x_gaussian_section_frequencies'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_frequencies'))

    x_gaussian_section_thermochem = SubSection(
        sub_section=SectionProxy('x_gaussian_section_thermochem'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_thermochem'))

    x_gaussian_section_force_constant_matrix = SubSection(
        sub_section=SectionProxy('x_gaussian_section_force_constant_matrix'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_force_constant_matrix'))

    x_gaussian_section_orbital_symmetries = SubSection(
        sub_section=SectionProxy('x_gaussian_section_orbital_symmetries'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_orbital_symmetries'))

    x_gaussian_section_symmetry = SubSection(
        sub_section=SectionProxy('x_gaussian_section_symmetry'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_symmetry'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_gaussian_energy_total_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total energy calculated with a given method during the self-consistent field (SCF)
        iterations.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_energy_total_scf_iteration'))

    x_gaussian_delta_energy_total_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total energy increment calculated with a given method during the self-consistent
        field (SCF) iterations.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_delta_energy_total_scf_iteration'))

    x_gaussian_single_configuration_calculation_converged = Quantity(
        type=str,
        shape=[],
        description='''
        Determines whether a single configuration calculation is converged.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_single_configuration_calculation_converged'))

    x_gaussian_spin_S2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Real value of spin squared.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_spin_S2'))

    x_gaussian_after_annihilation_spin_S2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Real value of spin squared resulting from the annihilation of the first spin
        contaminant.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_after_annihilation_spin_S2'))

    x_gaussian_energy_scf = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the SCF total energy, either HF or DFT.
        ''',
        categories=[public.energy_value, public.energy_total_potential, public.energy_component],
        a_legacy=LegacyDefinition(name='x_gaussian_energy_scf'))

    x_gaussian_perturbation_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Value of the perturbation energy for double hybrids
        ''',
        categories=[public.energy_value, public.energy_total_potential, public.energy_component],
        a_legacy=LegacyDefinition(name='x_gaussian_perturbation_energy'))

    x_gaussian_hf_detect = Quantity(
        type=str,
        shape=[],
        description='''
        Determine if the SCF method is one of RHF, UHF, or ROHF
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_hf_detect'))

    x_gaussian_energy_electrostatic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total electrostatic energy (nuclei + electrons), defined consistently with
        calculation_method.
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_gaussian_energy_electrostatic'))

    x_gaussian_energy_error = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Error in the total energy. Defined consistently with XC_method.
        ''',
        categories=[public.error_estimate_contribution, public.energy_value],
        a_legacy=LegacyDefinition(name='x_gaussian_energy_error'))

    x_gaussian_electronic_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Self-consistent electronic kinetic energy as defined in XC_method.
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_gaussian_electronic_kinetic_energy'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    x_gaussian_alpha_occ_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for occupied alpha MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_occ_eigenvalues_values'))

    x_gaussian_alpha_vir_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for virtual alpha MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_vir_eigenvalues_values'))

    x_gaussian_beta_occ_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for occupied beta MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_occ_eigenvalues_values'))

    x_gaussian_beta_vir_eigenvalues_values = Quantity(
        type=str,
        shape=[],
        description='''
        values of eigenenergies for virtual beta MOs
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_vir_eigenvalues_values'))

    x_gaussian_alpha_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_alpha_mos'],
        description='''
        values of eigenenergies, alpha occ
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_eigenvalues'))

    x_gaussian_beta_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_beta_mos'],
        description='''
        values of eigenenergies, beta occ
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_eigenvalues'))

    x_gaussian_alpha_occupations = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_alpha_mos'],
        description='''
        values of eigenenergies, alpha occ
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_alpha_occupations'))

    x_gaussian_beta_occupations = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_beta_mos'],
        description='''
        values of eigenenergies, beta occ
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_beta_occupations'))

    x_gaussian_eigenvalues_occupation = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        description='''
        Occupation of the eigenstates whose (energy) eigenvalues are given in
        eigenvalues_values.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_eigenvalues_occupation'))

    x_gaussian_eigenvalues_values = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        unit='joule',
        description='''
        Values of the (electronic-energy) eigenvalues. Their occupations are given in
        eigenvalues_occupation.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_eigenvalues_values'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_gaussian_settings = Quantity(
        type=str,
        shape=[],
        description='''
        electronic structure method, basis set, etc.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_settings'))

    x_gaussian_settings_corrected = Quantity(
        type=str,
        shape=[],
        description='''
        electronic structure method, basis set, etc.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_settings_corrected'))

    x_gaussian_method = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the electronic structure method used for the
        final wavefunctions.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_method'))

    x_gaussian_xc = Quantity(
        type=str,
        shape=[],
        description='''
        String identifying in an unique way the XC method used for the final
        wavefunctions.
        ''',
        a_legacy=LegacyDefinition(name='x_gaussian_xc'))

    x_gaussian_section_elstruc_method = SubSection(
        sub_section=SectionProxy('x_gaussian_section_elstruc_method'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_gaussian_section_elstruc_method'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask import request, g
from flask_restplus import Resource, fields, abort
import re

from nomad import utils, processing as proc
from nomad.datamodel import Dataset
from nomad.metainfo.flask_extension import generate_flask_restplus_model
from nomad.doi import DOI

from .. import common
from .api import api
from .auth import authenticate
from .common import pagination_model, pagination_request_parser
from .repo import edit


ns = api.namespace(
    'datasets',
    description='Datasets allow to create sets of related data.')

dataset_model = generate_flask_restplus_model(api, Dataset.m_def)
dataset_list_model = api.model('DatasetList', {
    'pagination': fields.Nested(model=pagination_model, skip_none=True),
    'results': fields.List(fields.Nested(model=dataset_model, skip_none=True))
})

list_datasets_parser = pagination_request_parser.copy()
list_datasets_parser.add_argument('prefix', help='Only return dataset with names that start with prefix.')


@ns.route('/')
class DatasetListResource(Resource):
    @api.doc('list_datasets')
    @api.marshal_with(dataset_list_model, skip_none=True, code=200, description='Dateset send')
    @api.expect(list_datasets_parser)
    @authenticate(required=True)
    def get(self):
        ''' Retrieve a list of all datasets of the authenticated user. '''
        args = {
            key: value for key, value in list_datasets_parser.parse_args().items()
            if value is not None}

        page = args.get('page', 1)
        per_page = args.get('per_page', 10)
        prefix = args.get('prefix', '')

        query_params = dict(user_id=g.user.user_id)
        if prefix != '':
            query_params.update(name=re.compile('^%s.*' % prefix, re.IGNORECASE))

        result_query = Dataset.m_def.a_mongo.objects(**query_params)

        return dict(
            pagination=dict(total=result_query.count(), page=page, per_page=per_page),
            results=[
                Dataset.m_def.a_mongo.to_metainfo(result)
                for result in result_query[(page - 1) * per_page: page * per_page]]), 200

    @api.doc('create_dataset')
    @api.response(400, 'The provided data is malformed or a dataset with the name already exists')
    @api.marshal_with(dataset_model, skip_none=True, code=200, description='Dateset send')
    @api.expect(dataset_model)
    @authenticate(required=True)
    def put(self):
        ''' Creates a new dataset. '''
        data = request.get_json()
        if data is None:
            data = {}

        # unique name
        name = data.get('name', None)
        if name is None:
            abort(400, 'Must provide a dataset name.')

        if Dataset.m_def.a_mongo.objects(user_id=g.user.user_id, name=name).count() > 0:
            abort(400, 'A dataset with name %s does already exist for the current user.' % name)

        # only admin can set user or doi
        if any(key in data for key in ['user_id', 'doi', 'dataset_id']):
            if not g.user.is_admin():
                abort(400, 'The dataset contains information you are not allowed to set.')

        # no other keys
        if any(key not in Dataset.m_def.all_quantities for key in data):  # pylint: disable=all
            abort(400, 'The dataset contains unknown keys.')

        if 'user_id' not in data:
            data['user_id'] = g.user.user_id
        dataset_id = data.pop('dataset_id', utils.create_uuid())
        return Dataset(dataset_id=dataset_id, **data).a_mongo.create(), 200


@ns.route('/<path:name>')
@api.doc(params=dict(name='The name of the requested dataset.'))
class DatasetResource(Resource):
    @api.doc('get_dataset')
    @api.response(404, 'The dataset does not exist')
    @api.marshal_with(dataset_model, skip_none=True, code=200, description='Dateset send')
    @authenticate(required=True)
    def get(self, name: str):
        ''' Retrieve a dataset by name. '''
        try:
            result = Dataset.m_def.a_mongo.get(user_id=g.user.user_id, name=name)
        except KeyError:
            abort(404, 'Dataset with name %s does not exist for current user' % name)

        return result

    @api.doc('assign_doi')
    @api.response(404, 'The dataset does not exist')
    @api.response(400, 'The dataset already has a DOI')
    @api.marshal_with(dataset_model, skip_none=True, code=200, description='DOI assigned')
    @authenticate(required=True)
    def post(self, name: str):
        ''' Assign a DOI to the dataset. '''
        try:
            result = Dataset.m_def.a_mongo.get(user_id=g.user.user_id, name=name)
        except KeyError:
            abort(404, 'Dataset with name %s does not exist for current user' % name)

        if result.doi is not None:
            abort(400, 'Dataset with name %s already has a DOI' % name)

        # check if the DOI can be created
        published_values = proc.Calc.objects(metadata__datasets=result.dataset_id).distinct('metadata.published')

        if False in published_values:
            abort(400, 'Dataset must not contain non published entries.')

        if True not in published_values:
            abort(400, 'Dataset must not be empty.')

        # set the DOI
        doi = DOI.create(title='NOMAD dataset: %s' % result.name, user=g.user)
        doi.create_draft()
        doi.make_findable()

        result.doi = doi.doi

        result.a_mongo.save()
        if doi.state != 'findable':
            common.logger.warning(
                'doi was created, but is not findable', doi=doi.doi, doi_state=doi.state,
                dataset=result.dataset_id)

        # update all affected calcs in the search index
        edit(dict(dataset_id=result.dataset_id))

        return result

    @api.doc('delete_dataset')
    @api.response(404, 'The dataset does not exist')
    @api.response(400, 'The dataset has a DOI and cannot be deleted')
    @api.marshal_with(dataset_model, skip_none=True, code=200, description='Dateset deleted')
    @authenticate(required=True)
    def delete(self, name: str):
        ''' Delete the dataset. '''
        try:
            result = Dataset.m_def.a_mongo.get(user_id=g.user.user_id, name=name)
        except KeyError:
            abort(404, 'Dataset with name %s does not exist for current user' % name)

        if result.doi is not None and len(result.doi) > 0:
            if g.user.is_admin:
                DOI.objects(doi__in=result.doi).delete()
            else:
                abort(400, 'Dataset with name %s has a DOI and cannot be deleted' % name)

        # edit all affected entries
        edit(
            dict(dataset_id=result.dataset_id),
            {'__raw__': {'$pull': {'metadata.datasets': result.dataset_id}}})

        # delete the dataset
        result.a_mongo.delete()

        return result


@ns.route('/doi/<path:doi>')
class RepoPidResource(Resource):
    @api.doc('resolve_doi')
    @api.response(404, 'Dataset with DOI does not exist')
    @api.marshal_with(dataset_model, skip_none=True, code=200, description='DOI resolved')
    @authenticate()
    def get(self, doi: str):
        dataset_me = Dataset.m_def.a_mongo.objects(doi=doi).first()
        if dataset_me is None:
            abort(404, 'Dataset with DOI %s does not exist' % doi)

        return dataset_me, 200



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='fplo_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='fplo.nomadmetainfo.json'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_fplo_program_version_sub = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO sub version
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_program_version_sub'))

    x_fplo_program_compilation_options = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO compilation options
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_program_compilation_options'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_fplo_reciprocal_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[3, 3],
        unit='1 / meter',
        description='''
        Reciprocal Lattice vectors (in Cartesian coordinates). The first index runs over
        the $x,y,z$ Cartesian coordinates, and the second index runs over the 3 lattice
        vectors.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_fplo_reciprocal_cell'))

    x_fplo_atom_idx = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        FPLO-internal index for each atom
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_atom_idx'))

    x_fplo_atom_wyckoff_idx = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        Wyckoff position index of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_atom_wyckoff_idx'))

    x_fplo_atom_cpa_block = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms'],
        description='''
        CPA block of each atom
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_atom_cpa_block'))

    x_fplo_structure_type = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO structure type: Crystal/Molecule
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_structure_type'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_fplo_xc_functional_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FPLO number xc functional
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_xc_functional_number'))

    x_fplo_xc_functional = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of xc functional
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_xc_functional'))

    x_fplo_dft_plus_u_projection_type = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of DFT+U projection
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_projection_type'))

    x_fplo_dft_plus_u_functional = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO notation of DFT+U functional
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_functional'))


class section_dft_plus_u_orbital(common.section_dft_plus_u_orbital):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_dft_plus_u_orbital'))

    x_fplo_dft_plus_u_orbital_element = Quantity(
        type=str,
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: element
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_element'))

    x_fplo_dft_plus_u_orbital_species = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: species index
        ''',
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_species'))

    x_fplo_dft_plus_u_orbital_F0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F0
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_F0'))

    x_fplo_dft_plus_u_orbital_F2 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F2
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_F2'))

    x_fplo_dft_plus_u_orbital_F4 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F4
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_F4'))

    x_fplo_dft_plus_u_orbital_F6 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        FPLO: Atom/Orbital dependent DFT+U property: value F6
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='x_fplo_dft_plus_u_orbital_F6'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List
from abc import ABCMeta, abstractmethod
import re
import os
import os.path
from functools import lru_cache

from nomad import config
from nomad.datamodel import EntryArchive, UserProvidableMetadata, EntryMetadata


class Parser(metaclass=ABCMeta):
    '''
    Instances specify a parser. It allows to find *main files* from  given uploaded
    and extracted files. Further, allows to run the parser on those 'main files'.
    '''
    name = "parsers/parser"

    def __init__(self):
        self.domain = 'dft'

    @abstractmethod
    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        '''
        Checks if a file is a mainfile for the parsers.

        Arguments:
            filename: The filesystem path to the mainfile
            mime: The mimetype of the mainfile guessed with libmagic
            buffer: The first 2k of the mainfile contents
            compression: The compression of the mainfile ``[None, 'gz', 'bz2']``
        '''
        pass

    @abstractmethod
    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        '''
        Runs the parser on the given mainfile and populates the result in the given
        archive root_section. It allows to be run repeatedly for different mainfiles.

        Args:
            mainfile: A path to a mainfile that this parser can parse.
            archive: An instance of the section :class:`EntryArchive`. It might contain
                a ``section_metadata`` with information about the entry.
            logger: A optional logger
        '''
        pass

    @classmethod
    def main(cls, mainfile):
        archive = EntryArchive()
        archive.m_create(EntryMetadata)
        cls().parse(mainfile, archive)  # pylint: disable=no-value-for-parameter
        return archive


class BrokenParser(Parser):
    '''
    A parser implementation that just fails and is used to match mainfiles with known
    patterns of corruption.
    '''
    name = 'parsers/broken'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.code_name = 'currupted mainfile'
        self._patterns = [
            re.compile(r'^pid=[0-9]+'),  # some 'mainfile' contain list of log-kinda information with pids
            re.compile(r'^Can\'t open .* library:.*')  # probably bad code runs
        ]

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:

        if decoded_buffer is not None:
            for pattern in self._patterns:
                if pattern.search(decoded_buffer) is not None:
                    return True

        return False

    def parse(self, mainfile: str, archive, logger=None):
        raise Exception('Failed on purpose.')


class MatchingParser(Parser):
    '''
    A parser implementation that uses regular expressions to match mainfiles.

    Arguments:
        code_name: The name of the code or input format
        code_homepage: The homepage of the code or input format
        mainfile_mime_re: A regexp that is used to match against a files mime type
        mainfile_contents_re: A regexp that is used to match the first 1024 bytes of a
            potential mainfile.
        mainfile_name_re: A regexp that is used to match the paths of potential mainfiles
        mainfile_alternative: If True files are mainfile if no mainfile_name_re matching file
            is present in the same directory.
        domain: The domain that this parser should be used for. Default is 'dft'.
        supported_compressions: A list of [gz, bz2], if the parser supports compressed files
    '''
    def __init__(
            self, name: str, code_name: str, code_homepage: str = None,
            mainfile_contents_re: str = None,
            mainfile_binary_header: bytes = None,
            mainfile_binary_header_re: bytes = None,
            mainfile_mime_re: str = r'text/.*',
            mainfile_name_re: str = r'.*',
            mainfile_alternative: bool = False,
            domain='dft',
            supported_compressions: List[str] = []) -> None:

        super().__init__()
        self.name = name
        self.code_name = code_name
        self.code_homepage = code_homepage
        self.domain = domain
        self._mainfile_binary_header = mainfile_binary_header
        self._mainfile_mime_re = re.compile(mainfile_mime_re)
        self._mainfile_name_re = re.compile(mainfile_name_re)
        self._mainfile_alternative = mainfile_alternative
        # Assign private variable this way to avoid static check issue.
        if mainfile_contents_re is not None:
            self._mainfile_contents_re = re.compile(mainfile_contents_re)
        else:
            self._mainfile_contents_re = None
        if mainfile_binary_header_re is not None:
            self._mainfile_binary_header_re = re.compile(mainfile_binary_header_re)
        else:
            self._mainfile_binary_header_re = None
        self._supported_compressions = supported_compressions

        self._ls = lru_cache(maxsize=16)(lambda directory: os.listdir(directory))

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:

        if self._mainfile_binary_header is not None:
            if self._mainfile_binary_header not in buffer:
                return False

        if self._mainfile_binary_header_re is not None:
            if buffer is not None:
                if self._mainfile_binary_header_re.search(buffer) is None:
                    return False
            else:
                return False

        if self._mainfile_contents_re is not None:
            if decoded_buffer is not None:
                if self._mainfile_contents_re.search(decoded_buffer) is None:
                    return False
            else:
                return False

        if self._mainfile_mime_re.match(mime) is None:
            return False

        if compression is not None and compression not in self._supported_compressions:
            return False

        if self._mainfile_name_re.fullmatch(filename) is None:
            if not self._mainfile_alternative:
                return False

            directory = os.path.dirname(filename)
            for sibling in self._ls(directory):
                sibling = os.path.join(directory, sibling)
                sibling_is_mainfile = sibling != filename and \
                    self._mainfile_name_re.fullmatch(sibling) is not None and \
                    os.path.isfile(sibling)
                if sibling_is_mainfile:
                    return False

        return True

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        raise NotImplementedError()

    def __repr__(self):
        return self.name


# For backward compatibility
FairdiParser = MatchingParser


class ArchiveParser(MatchingParser):
    def __init__(self):
        super().__init__(
            name='parsers/archive',
            code_name=config.services.unavailable_value,
            domain=None,
            mainfile_mime_re=r'application/json',
            mainfile_name_re=r'.*(archive|metainfo)\.json',
            mainfile_contents_re=r'section_metadata|results')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None):
        import json
        with open(mainfile, 'rt') as f:
            archive_data = json.load(f)

        metadata_data = archive_data.get(EntryArchive.section_metadata.name, None)

        if metadata_data is not None:
            metadata = archive.section_metadata
            for key, value in metadata_data.items():
                if UserProvidableMetadata.m_def not in getattr(EntryMetadata, key).categories:
                    continue

                metadata.m_update_from_dict({key: value})
            del(archive_data[EntryArchive.section_metadata.name])

        archive.m_update_from_dict(archive_data)


class MissingParser(MatchingParser):
    '''
    A parser implementation that just fails and is used to match mainfiles with known
    patterns of corruption.
    '''
    name = "parsers/missing"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def parse(self, mainfile: str, archive: EntryArchive, logger=None):
        raise Exception('The code %s is not yet supported.' % self.code_name)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)

from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import common_dft


class x_fhi_vibes_method(MCategory):
    '''
    Parameters from vibes metadata belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_fhi_vibes_method'))


class x_fhi_vibes_section_calculator_parameters(MSection):
    '''
    Calculator parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_calculator_parameters'))

    x_fhi_vibes_calculator_parameters_xc = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_xc'))

    x_fhi_vibes_calculator_parameters_k_grid = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_k_grid'))

    x_fhi_vibes_calculator_parameters_sc_accuracy_rho = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_sc_accuracy_rho'))

    x_fhi_vibes_calculator_parameters_relativistic = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_relativistic'))

    x_fhi_vibes_calculator_parameters_compensate_multipole_errors = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_compensate_multipole_errors'))

    x_fhi_vibes_calculator_parameters_output_level = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_output_level'))

    x_fhi_vibes_calculator_parameters_compute_forces = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_compute_forces'))

    x_fhi_vibes_calculator_parameters_compute_heat_flux = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_compute_heat_flux'))

    x_fhi_vibes_calculator_parameters_use_pimd_wrapper = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_use_pimd_wrapper'))

    x_fhi_vibes_calculator_parameters_species_dir = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator_parameters_species_dir'))


class x_fhi_vibes_section_calculator(MSection):
    '''
    Calculator parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_calculator'))

    x_fhi_vibes_calculator = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_calculator'))

    x_fhi_vibes_section_calculator_parameters = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_calculator_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_calculator_parameters'))


class x_fhi_vibes_section_atoms(MSection):
    '''
    Calculator parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_atoms'))

    x_fhi_vibes_atoms_kind = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_kind'))

    x_fhi_vibes_atoms_natoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_natoms'))

    x_fhi_vibes_atoms_pbc = Quantity(
        type=bool,
        shape=[3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_pbc'))

    x_fhi_vibes_atoms_cell = Quantity(
        type=np.dtype(np.float64),
        unit='meter',
        shape=[3, 3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_cell'))

    x_fhi_vibes_atoms_positions = Quantity(
        type=np.dtype(np.float64),
        unit='meter',
        shape=['x_fhi_vibes_atoms_natoms', 3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_positions'))

    x_fhi_vibes_atoms_velocities = Quantity(
        type=np.dtype(np.float64),
        unit='meter / second',
        shape=['x_fhi_vibes_atoms_natoms', 3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_velocities'))

    x_fhi_vibes_atoms_symbols = Quantity(
        type=str,
        shape=['x_fhi_vibes_atoms_natoms'],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_symbols'))

    x_fhi_vibes_atoms_masses = Quantity(
        type=np.dtype(np.float64),
        unit='kilogram',
        shape=['x_fhi_vibes_atoms_natoms'],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_atoms_masses'))

    x_fhi_vibes_atoms_info = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_atoms_info'))


class x_fhi_vibes_section_MD(MSection):
    '''
    Molecular dynamics parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_MD'))

    x_fhi_vibes_MD_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_type'))

    x_fhi_vibes_MD_md_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_md_type'))

    x_fhi_vibes_MD_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_temperature'))

    x_fhi_vibes_MD_friction = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_friction'))

    x_fhi_vibes_MD_fix_cm = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_fix_cm'))

    x_fhi_vibes_MD_fs = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_fs'))

    x_fhi_vibes_MD_kB = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_kB'))

    x_fhi_vibes_MD_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_dt'))

    x_fhi_vibes_MD_kg = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_kg'))


class x_fhi_vibes_section_relaxation_kwargs(MSection):
    '''
    Relaxation kwargs
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_relaxation_kwargs'))

    x_fhi_vibes_relaxation_kwargs_maxstep = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_kwargs_maxstep'))

    x_fhi_vibes_relaxation_kwargs_restart = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_kwargs_restart'))


class x_fhi_vibes_section_relaxation(MSection):
    '''
    Relaxation parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_relaxation'))

    x_fhi_vibes_relaxation_type = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_type'))

    x_fhi_vibes_relaxation_optimizer = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_optimizer'))

    x_fhi_vibes_relaxation_maxstep = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_maxstep'))

    x_fhi_vibes_relaxation_driver = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_driver'))

    x_fhi_vibes_relaxation_fmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_fmax'))

    x_fhi_vibes_relaxation_unit_cell = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_unit_cell'))

    x_fhi_vibes_relaxation_fix_symmetry = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_fix_symmetry'))

    x_fhi_vibes_relaxation_hydrostatic_strain = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_hydrostatic_strain'))

    x_fhi_vibes_relaxation_constant_volume = Quantity(
        type=bool,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_constant_volume'))

    x_fhi_vibes_relaxation_scalar_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_scalar_pressure'))

    x_fhi_vibes_relaxation_decimals = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_decimals'))

    x_fhi_vibes_relaxation_symprec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_symprec'))

    x_fhi_vibes_relaxation_workdir = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_relaxation_workdir'))

    x_fhi_vibes_section_relaxation_kwargs = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_relaxation_kwargs'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_relaxation_kwargs'))


class x_fhi_vibes_section_phonopy(MSection):
    '''
    Phonony parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_phonony'))

    x_fhi_vibes_phonopy_version = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_phonopy_version'))

    x_fhi_vibes_phonopy_supercell_matrix = Quantity(
        type=np.dtype(np.int32),
        shape=[3, 3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_phonopy_supercell_matrix'))

    x_fhi_vibes_phonopy_symprec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_phonopy_symprec'))

    x_fhi_vibes_phonopy_displacement_dataset = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_phonopy_displacement_dataset'))

    x_fhi_vibes_section_phonopy_primitive = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_atoms'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_phonopy_primitive'))


class x_fhi_vibes_section_vibes(MSection):
    '''
    Vibes specifications
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_vibes'))

    x_fhi_vibes_version = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_version'))


class x_fhi_vibes_section_settings(MSection):
    '''
    Metadata settings
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_settings'))

    x_fhi_vibes_settings_common = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_settings_common'))

    x_fhi_vibes_setttings_machine = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_setttings_machine'))

    x_fhi_vibes_settings_calculator = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_settings_calculator'))

    x_fhi_vibes_settings_files = Quantity(
        type=JSON,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_settings_files'))


class x_fhi_vibes_section_metadata(MSection):
    '''
    Metadata
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata'))

    x_fhi_vibes_section_metadata_calculator = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_calculator'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_calculator'))

    x_fhi_vibes_section_metadata_MD = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_MD'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_MD'))

    x_fhi_vibes_section_metadata_relaxation = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_relaxation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_relaxation'))

    x_fhi_vibes_section_metadata_phonopy = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_phonopy'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_phonopy'))

    x_fhi_vibes_section_metadata_atoms = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_atoms'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_atoms'))

    x_fhi_vibes_section_metadata_settings = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_settings'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_settings'))

    x_fhi_vibes_section_metadata_vibes = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_vibes'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_metadata_vibes'))


class x_fhi_vibes_section_attributes(MSection):
    '''
    Dataset attributes
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_fhi_vibes_section_attributes'))

    x_fhi_vibes_attributes_name = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_name'))

    x_fhi_vibes_attributes_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_name'))

    x_fhi_vibes_attributes_natoms = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_natoms'))

    x_fhi_vibes_attributes_time_unit = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_time_unit'))

    x_fhi_vibes_attributes_timestep = Quantity(
        type=np.dtype(np.float64),
        unit='second',
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_timestep'))

    x_fhi_vibes_attributes_nsteps = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_nsteps'))

    x_fhi_vibes_attributes_symbols = Quantity(
        type=str,
        shape=['x_fhi_vibes_attributes_system_natoms'],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_symbols'))

    x_fhi_vibes_attributes_masses = Quantity(
        type=np.dtype(np.float64),
        unit='kilogram',
        shape=['x_fhi_vibes_attributes_system_natoms'],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_masses'))

    x_fhi_vibes_attributes_hash = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_hash'))

    x_fhi_vibes_attributes_hash_raw = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_hash_raw'))

    x_fhi_vibes_attributes_sigma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_sigma'))

    x_fhi_vibes_attributes_st_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_system_st_size'))

    x_fhi_vibes_attributes_number_of_atoms_primitive = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_number_of_atoms_primitive'))

    x_fhi_vibes_attributes_number_of_atoms_supercell = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_number_of_atoms_supercell'))

    x_fhi_vibes_attributes_force_constants_remapped_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_attributes_force_constants_remapped_size'))

    x_fhi_vibes_section_attributes_atoms = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_atoms'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_attributes_atoms'))

    x_fhi_vibes_section_attributes_metadata = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_metadata'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_attributes_metadata'))


class section_method(common_dft.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_fhi_vibes_section_attributes = SubSection(
        sub_section=SectionProxy('x_fhi_vibes_section_attributes'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_fhi_vibes_section_attributes'))


class section_single_configuration_calculation(common_dft.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_fhi_vibes_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter**3',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_volume'))

    x_fhi_vibes_displacements = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3, 3],
        unit='meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_displacements'))

    x_fhi_vibes_momenta = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3, 3],
        unit='kilogram * meter / second',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_momenta'))

    x_fhi_vibes_force_constants = Quantity(
        type=np.dtype(np.float64),
        shape=['x_fhi_vibes_attributes_number_of_atoms_primitive', 'x_fhi_vibes_attributes_number_of_atoms_primitive', 3, 3],
        unit='newton / meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_force_constants'))

    x_fhi_vibes_force_constants_remapped = Quantity(
        type=np.dtype(np.float64),
        shape=['x_fhi_vibes_attributes_force_constants_remapped_size', 'x_fhi_vibes_attributes_force_constants_remapped_size'],
        unit='newton / meter',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_force_constants'))

    x_fhi_vibes_atom_forces_harmonic = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='newton',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_forces_harmonic'))

    x_fhi_vibes_energy_potential_harmonic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_energy_potential_harmonic'))

    x_fhi_vibes_heat_flux = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        unit='kilogram / second**3',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_heat_flux'))

    x_fhi_vibes_heat_flux_harmonic = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        unit='kilogram / second**3',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_heat_flux_harmonic'))

    x_fhi_vibes_heat_flux_0_harmonic = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        unit='kilogram / second**3',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_heat_flux_0_harmonic'))

    x_fhi_vibes_sigma_per_sample = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_sigma_per_sample'))

    x_fhi_vibes_pressure_kinetic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_pressure_kinetic'))

    x_fhi_vibes_pressure_potential = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_pressure_potential'))

    # TODO in fhi aims this is raw id, however in vibes, where each calculation corresponds
    # to an scc, we get a list of these ids, it does not make sense to put these in sec_run
    x_fhi_vibes_aims_uuid = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_aims_uuid'))

    x_fhi_vibes_MD_time = Quantity(
        type=np.dtype(np.float64),
        unit='second',
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_fhi_vibes_MD_time'))
